program ShellExpl;

uses Windows, Messages, CommCtrl, ShlObj,
  ShellAPI,
  ActiveX, UtilsUnit;//, SysUtils;

{$R RESOURCE.RES}

const
  szClassName = 'ANS-GenericClass';
  szCaption   = 'Generic Window';
  IDI_MAINICON = 101;

var
  hwndMain: HWND;
  hwndLV, hwndTV: HWND;
  
(*
#if (_WIN32_WINNT >= 0x501)
#define LVS_EX_DOUBLEBUFFER     0x00010000
#define LVS_EX_HIDELABELS       0x00020000
#define LVS_EX_SINGLEROW        0x00040000
#define LVS_EX_SNAPTOGRID       0x00080000  // Icons automatically snap to grid.
#define LVS_EX_SIMPLESELECT     0x00100000  // Also changes overlay rendering to top right for icon mode.
#endif
*)

const
  LVS_EX_DOUBLEBUFFER     = $00010000;
  LVS_EX_HIDELABELS       = $00020000;
  LVS_EX_SINGLEROW        = $00040000;
  LVS_EX_SNAPTOGRID       = $00080000;
  LVS_EX_SIMPLESELECT     = $00100000;

const
  SP_WIDTH    = 3;
  SC_DRAGMOVE = $F012;

var
  TVWidth: Integer = 300;
  
const
  MIN_VALUE = 100;
  MAX_VALUE = 600;
  
type
  PShellFolder = ^TShellFolder;
  TShellFolder = packed record
    FPIDL,
    FFullPIDL: PITEMIDLIST;
    FParentFolder: IShellFolder;
    FDisplayName: string;
    FIconIndex: Integer;
  end;

  TFolder = class
  private
    FPIDL,
    FFullPIDL: PITEMIDLIST;
    FParentFolder: IShellFolder;
    FDisplayName: string;
    FIconIndex: Integer;
    FSelectedIndex: Integer;
  protected
    function GetDisplayName: string;
    function GetIconIndex: Integer;
  public
    constructor Create(ParentFolder: IShellFolder; PIDL, FullPIDL: PITEMIDLIST);
    destructor Destroy; override;
    property DisplayName: string read FDisplayName;
    property IconIndex: Integer read FIconIndex;
    property SelectedIndex: Integer read FSelectedIndex;
    //property DisplayName: string read GetDisplayName;
    //property IconIndex: Integer read GetIconIndex;
  end;

var
  FolderList: TList;
  TreeList: TList;
  FIDesktopFolder: IShellFolder;
  FDesktopPIDL: PITEMIDLIST;
  Malloc: IMalloc;
  System_ImageList_Large, System_ImageList_Small: HIMAGELIST;
  hwndIconBtn: HWND;
  CurrCSIDL: Integer = CSIDL_DESKTOP;
  CurrIcon: HICON = 0;
  CurrPath: string;
  IsDesktopRoot: Boolean = True;

{ TFolder }

constructor TFolder.Create(ParentFolder: IShellFolder; PIDL, FullPIDL: PITEMIDLIST);
begin
  FParentFolder := ParentFolder;
  FPIDL := PIDL;
  FFullPIDL := FullPIDL;
  if Assigned(FParentFolder) then
    FDisplayName := UtilsUnit.GetDisplayName(FParentFolder, FPIDL, SHGDN_INFOLDER)
  else
    FDisplayName := UtilsUnit.GetDisplayName(FIDesktopFolder, FPIDL, SHGDN_INFOLDER);
  FIconIndex := UtilsUnit.GetShellImage(FFullPIDL, True, False);
  FSelectedIndex := FIconIndex;
  //FSelectedIndex := UtilsUnit.GetShellImage(FFullPIDL, True, True);
end;

function TFolder.GetDisplayName: string;
begin
  Result := UtilsUnit.GetDisplayName(FIDesktopFolder, FFullPIDL, SHGDN_INFOLDER);
end;

function TFolder.GetIconIndex: Integer;
begin
  Result := UtilsUnit.GetShellImage(FFullPIDL, True, False);
end;

destructor TFolder.Destroy;
begin
  if Assigned(FPIDL) then
    Malloc.Free(FPIDL);
  if Assigned(FFullPIDL) then
    Malloc.Free(FFullPIDL);
end;

function ListViewCompareFunc(Item1, Item2: Pointer): Integer;
var
  Folder1, Folder2: TFolder;
  ShellFolder: IShellFolder;
begin
  Result := 0;
  if (Item1 = nil) or (Item2 = nil) then Exit; 
  Folder1 := TFolder(Item1);
  Folder2 := TFolder(Item2);
  ShellFolder := Folder1.FParentFolder;
  if Assigned(ShellFolder) then
    Result := SmallInt(ShellFolder.CompareIDs(0, Folder1.FPIDL, Folder2.FPIDL));
end;

function UpdateDestopFolder(hDlg: HWND): LRESULT;

const
  Flags = SHCONTF_FOLDERS or SHCONTF_NONFOLDERS;// or SHCONTF_INCLUDEHIDDEN;

var
  ID: PItemIDList;
  EnumList: IEnumIDList;
  NumIDs: LongWord;
  Folder: TFolder;
  Count: Integer;

begin

  Result := 0;

  //FolderList.Clear;
  //FolderList.Pack;

  Count := 0;

  if FDesktopPIDL = nil then Exit;

  try
    OleCheck(
      FIDesktopFolder.EnumObjects(
        hDlg,
        Flags,
        EnumList)
    );

    if not Assigned(EnumList) then
      Exit;

    while EnumList.Next(1, ID, NumIDs) = S_OK do
    begin

      Folder := TFolder.Create(FIDesktopFolder, ID,
        ConcatPIDLs(FDesktopPIDL, ID));

      FolderList.Add(Folder);

      Inc(Count);

    end;

  finally

  end;

  FolderList.Sort(@ListViewCompareFunc);

  ListView_SetItemCount(hwndLV, Count);

end;

function TreeViewSortFunc(lParam1, lParam2, lParamSort: LPARAM): Integer; stdcall;
var
  PFolder1, PFolder2: PShellFolder;
  ShellFolder: IShellFolder;
begin
  Result := 0;
  PFolder1 := PShellFolder(lParam1);
  PFolder2 := PShellFolder(lParam2);
  if Assigned(PFolder1) and Assigned(PFolder2) then
  begin
    ShellFolder := PFolder1^.FParentFolder;
    if Assigned(ShellFolder) then
      Result := SmallInt(ShellFolder.CompareIDs(0, PFolder1^.FPIDL, PFolder2^.FPIDL));
  end;
end;

function UpdateTreeViewDesktop(hDlg: HWND): LRESULT;

const
  Flags = SHCONTF_FOLDERS;

var

  DisplayName: string;
  Param: PShellFolder;

var
  tvis: TTVINSERTSTRUCT;
  Parent: HTREEITEM;

begin

  Result := 0;

  TreeView_DeleteAllItems(hwndTV);

  FillChar(tvis, Sizeof(TTVINSERTSTRUCT), 0);

  if FDesktopPIDL = nil then Exit;

  DisplayName := GetDisplayName(FIDesktopFolder, FDesktopPIDL, SHGDN_INFOLDER);

  FillChar(tvis, SizeOf(tvis), #0);

  New(Param);
  Param^.FPIDL := CopyPIDL(FDesktopPIDL);
  Param^.FFullPIDL := CopyPIDL(FDesktopPIDL);
  Param^.FParentFolder := nil;

  with tvis do
  begin
    hParent := TVI_ROOT;
    hInsertAfter := TVI_LAST;
    with item do
    begin
      mask := TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or
              TVIF_PARAM or TVIF_CHILDREN;
      pszText := PChar(DisplayName);
      cchTextMax := Length(DisplayName);
      iImage := GetShellImage(FDesktopPIDL, False, False);
      iSelectedImage := GetShellImage(FDesktopPIDL, False, True);
      if GetHasSubFolders(FIDesktopFolder, FDesktopPIDL) then
        cChildren := 1;
      lParam := LongInt(Param);
    end;
  end;

  Parent := TreeView_InsertItem(hwndTV, tvis);

  TreeView_SelectItem(hwndTV, Parent);
  TreeView_Expand(hwndTV, Parent, TVE_EXPAND);

end;

function UpdateTreeView(hDlg: HWND): LRESULT;

const
  Flags = SHCONTF_FOLDERS or SHCONTF_INCLUDEHIDDEN;

var
  FRootPIDL: PITEMIDLIST;
  DisplayName: string;
  Param: PShellFolder;

var
  tvis: TTVINSERTSTRUCT;
  Parent: HTREEITEM;

begin

  Result := 0;

  if CurrCSIDL = CSIDL_DESKTOP then
  begin
    UpdateTreeViewDesktop(hDlg);
    Exit;
  end;

  TreeView_DeleteAllItems(hwndTV);

  FillChar(tvis, Sizeof(TTVINSERTSTRUCT), 0);

  OLECheck(
    SHGetSpecialFolderLocation(
      hDlg,
      CurrCSIDL,
      FRootPIDL)
  );

  if FRootPIDL = nil then Exit;

  DisplayName := GetDisplayName(FIDesktopFolder, FRootPIDL, SHGDN_INFOLDER);

  FillChar(tvis, SizeOf(tvis), #0);

  New(Param);
  Param^.FPIDL := FRootPIDL;
  Param^.FFullPIDL := CopyPIDL(FRootPIDL);
  Param^.FParentFolder := FIDesktopFolder;

  with tvis do
  begin
    hParent := TVI_ROOT;
    hInsertAfter := TVI_LAST;
    with item do
    begin
      mask := TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or
              TVIF_PARAM or TVIF_CHILDREN;
      pszText := PChar(DisplayName);
      cchTextMax := Length(DisplayName);
      iImage := GetShellImage(FRootPIDL, False, False);
      iSelectedImage := GetShellImage(FRootPIDL, False, True);
      if GetHasSubFolders(FIDesktopFolder, FRootPIDL) then
        cChildren := 1;
      lParam := LongInt(Param);
    end;
  end;

  Parent := TreeView_InsertItem(hwndTV, tvis);

  TreeView_Expand(hwndTV, Parent, TVE_EXPAND);
  TreeView_SelectItem(hwndTV, Parent);

end;

function UpdatePathListView(hDlg: HWND): Integer;

const
  Flags = SHCONTF_FOLDERS or SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN;
var
  P: PWideChar;
  IFlags,
  NumChars: LongWord;

var
  ID: PItemIDList;
  FIRootFolder: IShellFolder;
  FRootPIDL: PITEMIDLIST;
  EnumList: IEnumIDList;
  NumIDs: LongWord;
  Folder: TFolder;
  Count: Integer;

begin

  Result := 0;
  
  Count := 0;

  FolderList.Clear;

  //FolderList.Pack;

  ListView_DeleteAllItems(hwndLV);

  if IsDesktopRoot and (CurrCSIDL = CSIDL_DESKTOP) then
  begin
    UpdateDestopFolder(hDlg);
    Exit;
  end;

  NumChars := Length(CurrPath);
  IFlags := 0;
  P := StringToOleStr(CurrPath);
  try
    OLECheck(FIDesktopFolder.ParseDisplayName(
          hDlg,
          nil,
          P,
          NumChars,
          FRootPIDL,
          IFlags)
       );
  except

  end;

  if FRootPIDL = nil then
  begin
    //SetWindowText(GetDlgItem(hDlg, IDC_STATIC), PChar('Folder is nil.'));
    Exit;
  end;

  OLECheck(
     FIDesktopFolder.BindToObject(
            FRootPIDL,
            nil,
            IID_IShellFolder,
            Pointer(FIRootFolder))
   );

  try
    OleCheck(
      FIRootFolder.EnumObjects(
        hDlg,
        Flags,
        EnumList)
    );

    if not Assigned(EnumList) then
      Exit;

    while EnumList.Next(1, ID, NumIDs) = S_OK do
    begin

      Folder := TFolder.Create(FIRootFolder, ID, ConcatPIDLs(FRootPIDL, ID));

      FolderList.Add(Folder);

      Inc(Count);

    end;

  finally
  end;

  FolderList.Sort(@ListViewCompareFunc);

  ListView_SetItemCount(hwndLV, Count);

end;


function ExpandTreeItem(hDlg: HWND; Node: HTREEITEM; Folder: PShellFolder): LRESULT;
const
  Flags = SHCONTF_FOLDERS or SHCONTF_INCLUDEHIDDEN;

var
  AbsoluteID, ID: PItemIDList;
  FIRootFolder: IShellFolder;
  FRootPIDL: PITEMIDLIST;
  EnumList: IEnumIDList;
  NumIDs: LongWord;
  DisplayName: string;
  //IconIndex: Integer;
  Param: PShellFolder;

var
  tvis: TTVINSERTSTRUCT;
  Sort: tagTVSORTCB;

begin
  Result := 0;

  if (not Assigned(Node)) or (not Assigned(Folder)) then
    Exit;

  if not Assigned(Folder^.FParentFolder) then
  begin
  try

    OleCheck(
      FIDesktopFolder.EnumObjects(
        hDlg,
        Flags,
        EnumList)
    );

    if not Assigned(EnumList) then
      Exit;

    while EnumList.Next(1, ID, NumIDs) = S_OK do
    begin

      FillChar(tvis, SizeOf(tvis), #0);

      DisplayName := GetDisplayName(FIDesktopFolder, ID, SHGDN_INFOLDER);

      New(Param);
      Param^.FPIDL := ID;
      Param^.FFullPIDL := ConcatPIDLs(FDesktopPIDL, ID);
      Param^.FParentFolder := FIDesktopFolder;

      with tvis do
      begin
        hParent := Node;
        hInsertAfter := TVI_LAST;
        with item do
        begin
          mask := TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or
                  TVIF_PARAM or TVIF_CHILDREN;
          pszText := PChar(DisplayName);
          cchTextMax := Length(DisplayName);
          iImage := GetShellImage(ID, False, False);
          iSelectedImage := GetShellImage(ID, False, True);
          if iSelectedImage = 0 then iSelectedImage := iImage;
          if GetHasSubFolders(FIDesktopFolder, ID) then
            cChildren := 1;
          lParam := LongInt(Param);
        end;
        TreeView_InsertItem(hwndTV, tvis);

      end;

    end;

  finally

  end; // end try

  Sort.hParent := Node;
  Sort.lpfnCompare := @TreeViewSortFunc;
  Sort.lParam := LongInt(FIDesktopFolder);
  TreeView_SortChildrenCB(hwndTV, Sort, 0);

  end // end if
  else
  begin
  if not Assigned(Folder^.FPIDL) then Exit;

  FRootPIDL := Folder^.FFullPIDL;

  OLECheck(
     FIDesktopFolder.BindToObject(
            FRootPIDL,
            nil,
            IID_IShellFolder,
            Pointer(FIRootFolder))
   );

  try
    OleCheck(
      FIRootFolder.EnumObjects(
        hDlg,
        Flags,
        EnumList)
    );

    if not Assigned(EnumList) then
      Exit;

    while EnumList.Next(1, ID, NumIDs) = S_OK do
    begin

      AbsoluteID := ConcatPIDLs(FRootPIDL, ID);
      DisplayName := GetDisplayName(FIRootFolder, ID, SHGDN_INFOLDER);
      //IconIndex := GetShellImage(AbsoluteID, False, False);
      FillChar(tvis, SizeOf(tvis), #0);

      New(Param);
      Param^.FPIDL := ID;
      Param^.FFullPIDL := AbsoluteID;
      Param^.FParentFolder := FIRootFolder;

      with tvis do
      begin
        hParent := Node;
        hInsertAfter := TVI_LAST;
        with item do
        begin
          mask := TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or
                  TVIF_PARAM or TVIF_CHILDREN;
          pszText := PChar(DisplayName);
          cchTextMax := Length(DisplayName);
          iImage := GetShellImage(AbsoluteID, False, False);
          iSelectedImage := GetShellImage(AbsoluteID, False, True);
          if iSelectedImage = 0 then iSelectedImage := iImage;
          if GetHasSubFolders(FIRootFolder, ID) then
            cChildren := 1;
          lParam := LongInt(Param);
        end;
        TreeView_InsertItem(hwndTV, tvis);
      end;

    end;

  finally

  end; // end try

  Sort.hParent := Node;
  Sort.lpfnCompare := @TreeViewSortFunc;
  Sort.lParam := LongInt(FIDesktopFolder);
  TreeView_SortChildrenCB(hwndTV, Sort, 0);

  end; // end else

end;

{ InitTreeView }
function InitTreeView(Wnd: HWND): HWND;
begin

  Result := CreateWindowEx(WS_EX_CLIENTEDGE,
    WC_TREEVIEW,
    nil,
    WS_CHILD or WS_VISIBLE or TVS_HASBUTTONS or TVS_HASLINES or
    TVS_LINESATROOT or TVS_SHOWSELALWAYS,
    0, 0, 0, 0, Wnd, 0, hInstance, nil);

end;

var
  PrevLVWndProc: LongInt;

function ListViewWndProc(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT; stdcall;
begin

  Result := 0;

  case uMsg of
    WM_ERASEBKGND:
      Exit;
  else
    Result := CallWindowProc(TFNWndProc(PrevLVWndProc), Wnd, uMsg, wParam, lParam);
  end;

end;

{ InitListView }
function InitListView(Wnd: HWND): HWND;
var
  Style: Integer;
begin

  Result := CreateWindowEx(WS_EX_CLIENTEDGE,
    WC_LISTVIEW,
    nil,
    WS_CHILD or WS_VISIBLE or LVS_ICON or LVS_OWNERDATA,
    0, 0, 0, 0, Wnd, 0, hInstance, nil);

  Style := GetClassLong(Result, GWL_STYLE);
  Style := Style and not (CS_HREDRAW or CS_VREDRAW);
  SetClassLong(Result, GWL_STYLE, Style);

  //PrevLVWndProc := SetWindowLong(Result, GWL_WNDPROC, LongInt(@ListViewWndProc));

  ListView_SetExtendedListViewStyle(Result, LVS_EX_DOUBLEBUFFER);

end;

{ OnCreate }
function OnCreate(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  FileInfo: TShFileInfo;

begin

  Result := 0;

  Application := Wnd;

  hwndMain := Wnd;

  FolderList := TList.Create;

  TreeList := TList.Create;

  hwndTV := InitTreeView(Wnd);

  hwndLV := InitListView(Wnd);

  System_ImageList_Large := SHGetFileInfo('C:\',
    0,
    FileInfo,
    SizeOf(FileInfo),
    SHGFI_SYSICONINDEX or SHGFI_LARGEICON);

  System_ImageList_Small := SHGetFileInfo('C:\',
    0,
    FileInfo,
    SizeOf(FileInfo),
    SHGFI_SYSICONINDEX or SHGFI_SMALLICON);

  ListView_SetImageList(hwndLV, System_ImageList_Large, LVSIL_NORMAL);

  ListView_SetImageList(hwndLV, System_ImageList_Small, LVSIL_SMALL);

  TreeView_SetImageList(hwndTV, System_ImageList_Small, TVSIL_NORMAL);

  OLECheck(SHGetMalloc(Malloc));

  OLECheck(SHGetDesktopFolder(FIDesktopFolder));

  OLECheck(
    SHGetSpecialFolderLocation(
      Wnd,
      CSIDL_DESKTOP,
      FDesktopPIDL)
  );

  CurrPath := GetDisplayName(FIDesktopFolder, FDesktopPIDL, SHGDN_INFOLDER);

  UpdateTreeView(Wnd);

end;

function OnTreeViewDeleteItem(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMTREEVIEW;
  PFolder: PShellFolder;
begin
  Result := 0;
  pnmv := PNMTREEVIEW(lParam);
  PFolder := PShellFolder(pnmv^.itemOld.lParam);
  if Assigned(PFolder) then
  begin
    if Assigned(PFolder^.FPIDL) then
      Malloc.Free(PFolder^.FPIDL);
    if Assigned(PFolder^.FFullPIDL) then
      Malloc.Free(PFolder^.FFullPIDL);
    Dispose(PFolder);
  end;
end;

function OnTreeViewItemExpanded(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMTREEVIEW;
  PFolder: PShellFolder;
  DisplayName: string;
begin
  Result := 0;
  DisplayName := '';
  pnmv := PNMTREEVIEW(lParam);
  PFolder := PShellFolder(pnmv^.itemNew.lParam);
  if Assigned(PFolder) then
  begin
    if Assigned(PFolder^.FPIDL) then
      DisplayName := GetDisplayName(FIDesktopFolder, PFolder^.FPIDL, SHGDN_INFOLDER);
  end;
  case pnmv^.action of
    TVE_COLLAPSE:; //MessageBox(hDlg, 'TVE_COLLAPSE', PChar(DisplayName), MB_OK);
    TVE_EXPAND:
      if Assigned(PFolder) then
      begin
        if TreeView_GetChild(hwndTV, pnmv^.itemNew.hItem) = nil then
          ExpandTreeItem(hDlg, pnmv^.itemNew.hItem, PFolder);
      end;
      //MessageBox(hDlg, 'TVE_EXPAND', PChar(DisplayName), MB_OK);
    TVE_COLLAPSERESET: MessageBox(hDlg, 'TVE_COLLAPSERESET', 'Expanded', MB_OK);
    TVE_EXPANDPARTIAL: MessageBox(hDlg, 'TVE_EXPANDPARTIAL', 'Expanded', MB_OK);
    TVE_TOGGLE: MessageBox(hDlg, 'TVE_TOGGLE', 'Expanded', MB_OK);
  else
    MessageBox(hDlg, 'UNKNOWN', 'Expanded', MB_OK);
  end;
end;

function OnTreeViewSelChanged(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMTREEVIEW;
  PFolder: PShellFolder;
  DisplayName: string;
  //IconIndex: Integer;
  ParentFolder: IShellFolder;
begin
  Result := 0;
  pnmv := PNMTREEVIEW(lParam);
  PFolder := PShellFolder(pnmv^.itemNew.lParam);
  ParentFolder := PFolder^.FParentFolder;
  IsDesktopRoot := not Assigned(ParentFolder);
  if not Assigned(ParentFolder) then ParentFolder := FIDesktopFolder;
  if Assigned(PFolder) then
  begin
    if Assigned(PFolder^.FPIDL) and (Assigned(PFolder^.FFullPIDL)) then
    begin

      CurrPath := GetDisplayNameOf(ParentFolder, PFolder^.FPIDL, True);
      //CurrPath := GetDisplayName(ParentFolder, PFolder^.FPIDL, SHGDN_FORPARSING);

      DisplayName := GetDisplayName(FIDesktopFolder, PFolder^.FFullPIDL, SHGDN_INFOLDER);//SHGDN_FORPARSING);
      //DisplayName := GetDisplayName(ParentFolder, PFolder^.FPIDL, SHGDN_INFOLDER);
      //IconIndex := GetShellImage(PFolder^.FFullPIDL, True, False);
      if CurrIcon <> 0 then DestroyIcon(CurrIcon);
      //CurrIcon := ImageList_GetIcon(System_ImageList_Large, IconIndex, ILD_TRANSPARENT);
      CurrIcon := GetShellImageIcon(PFolder^.FFullPIDL, True, False);
      SendMessage(GetDlgItem(hDlg, 1004), BM_SETIMAGE, IMAGE_ICON, CurrIcon);
      //SetWindowText(GetDlgItem(hDlg, IDC_STATIC), PChar(DisplayName));

      //DisplayName := GetDisplayName(ParentFolder, PFolder^.FPIDL, SHGDN_FORPARSING);//SHGDN_FORPARSING);
      DisplayName := CurrPath;
      //SetWindowText(GetDlgItem(hDlg, IDC_EDIT), PChar(DisplayName));

      UpdatePathListView(hDlg);

    end;
  end;
end;

function OnNotifyTreeView(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMTREEVIEW;
begin
  Result := 0;
  pnmv := PNMTREEVIEW(lParam);
  case pnmv^.hdr.code of
    TVN_DELETEITEM:
      Result := OnTreeViewDeleteItem(hDlg, uMsg, wParam, lParam);
    TVN_ITEMEXPANDING:
      Result := OnTreeViewItemExpanded(hDlg, uMsg, wParam, lParam);
    TVN_SELCHANGED:
      Result := OnTreeViewSelChanged(hDlg, uMsg, wParam, lParam);
  end;
end;

function OnListViewItemChanged(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMLISTVIEW;
  Folder: TFolder;
  ParentFolder: IShellFolder;
  DisplayName: string;
begin
  Result := 0;
  pnmv := PNMLISTVIEW(lParam);
  if (pnmv^.iItem < 0) or (pnmv^.iItem > FolderList.Count) then Exit;
  Folder := TFolder(FolderList[pnmv^.iItem]);
  if not Assigned(Folder) then Exit;
  ParentFolder := Folder.FParentFolder;
  if not Assigned(ParentFolder) then ParentFolder := FIDesktopFolder;
  if Assigned(Folder) then
  begin
    if Assigned(Folder.FPIDL) and (Assigned(Folder.FFullPIDL)) then
    begin
      if CurrIcon <> 0 then DestroyIcon(CurrIcon);
      CurrIcon := GetShellImageIcon(Folder.FFullPIDL, True, False);
      SendMessage(hwndIconBtn, BM_SETIMAGE, IMAGE_ICON, CurrIcon);
      DisplayName := Folder.DisplayName;
      //SetWindowText(GetDlgItem(hDlg, IDC_STATIC), PChar(DisplayName));
      //SetWindowText(GetDlgItem(hDlg, IDC_EDIT), PChar(DisplayName));
    end;
  end;
end;

function OnListViewGetDispInfo(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PLVDISPINFO;
  Folder: TFolder;
  //ParentFolder: IShellFolder;
  //DisplayName: string;
begin

  Result := 0;

  pnmv := PLVDISPINFO(lParam);

  with pnmv^ do
  begin

    if (item.iItem < 0) or (item.iItem > FolderList.Count) then Exit;

    Folder := TFolder(FolderList[item.iItem]);
    if not Assigned(Folder) then Exit;

    if item.mask and LVIF_IMAGE <> 0 then
      item.iImage := Folder.IconIndex;

    if item.mask and LVIF_TEXT <> 0 then
    case item.iSubItem of
      //0: StrCopy(item.pszText, PChar(Folder.DisplayName));
      0: lstrcpy(item.pszText, PChar(Folder.DisplayName));
    end;

  end;
  
end;

function OnNotifyListView(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmv: PNMLISTVIEW;
begin
  Result := 0;
  pnmv := PNMLISTVIEW(lParam);
  case pnmv^.hdr.code of
    NM_RCLICK:;
      //Result := OnListViewRClick(hDlg, uMsg, wParam, lParam);
    LVN_DELETEALLITEMS:;
      //Result := OnListViewDeleteAllItems(hDlg, uMsg, wParam, lParam);
    LVN_DELETEITEM:;
      //Result := OnListViewDeleteItem(hDlg, uMsg, wParam, lParam);
    LVN_ITEMCHANGED:
      Result := OnListViewItemChanged(hDlg, uMsg, wParam, lParam);
    LVN_GETDISPINFO:
      Result := OnListViewGetDispInfo(hDlg, uMsg, wParam, lParam);
  end;
end;

function OnNotify(hDlg: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  pnmh: PNMHDR;
begin
  Result := 0;
  pnmh := PNMHDR(lParam);
  if pnmh^.hwndFrom = hwndLV then
    Result := OnNotifyListView(hDlg, uMsg, wParam, lParam)
  else if pnmh^.hwndFrom = hwndTV then
    Result := OnNotifyTreeView(hDlg, uMsg, wParam, lParam);
  {
  case pnmh^.idFrom of
    IDC_LISTVIEW:
      Result := OnNotifyListView(hDlg, uMsg, wParam, lParam);
    IDC_TREEVIEW:
      Result := OnNotifyTreeView(hDlg, uMsg, wParam, lParam);
  end;
  }
end;

{ OnDestroy }
function OnDestroy(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  Index: Integer;
  Folder: TFolder;
begin
  Result := 0;
  for Index := 0 to FolderList.Count-1 do
  begin
    Folder := TFolder(FolderList[Index]);
    if Assigned(Folder) then
      Folder.Free;
  end;
  FolderList.Free;

  for Index := 0 to TreeList.Count-1 do
  begin
    Folder := TFolder(TreeList[Index]);
    if Assigned(Folder) then
      Folder.Free;
  end;
  TreeList.Free;

  if CurrIcon <> 0 then DestroyIcon(CurrIcon);

  ListView_DeleteAllItems(hwndLV);
  TreeView_DeleteAllItems(hwndTV);
end;

procedure GetClientView(Wnd: HWND; var R: TRect);
begin
  GetClientRect(Wnd, R);
end;

function EnumFunc(hWndChild: HWND; lParam: LongInt): BOOL stdcall;
var
  //R: PRect;
  CR, TR: TRect;
begin
  //R := PRect(lParam);
  if hWndChild = hwndTV then
  begin
    GetClientView(GetParent(hWndChild), CR);
    MoveWindow(hWndChild, CR.Left, CR.Top,
      TVWidth, (CR.Bottom - CR.Top), True);
  end
  else if hWndChild = hwndLV then
  begin
    GetClientView(GetParent(hWndChild), CR);
    GetWindowRect(hWndTV, TR);
    MoveWindow(hWndChild, CR.Left + (TR.Right - TR.Left) + SP_WIDTH, CR.Top,
      (CR.Right - CR.Left) - (TR.Right - TR.Left) - SP_WIDTH, (TR.Bottom - TR.Top), True);
    InvalidateRect(hwndLV, nil, True);
  end;
  Result := True;
end;

{ OnSize }
function OnSize(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  ClientRect: TRect;
begin
  Result := 0;
  GetClientRect(Wnd, ClientRect);
  EnumChildWindows(Wnd, @EnumFunc, LongInt(@ClientRect));
end;

{ OnClose }
function OnClose(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
begin
  Result := 0;
end;

function GetSplitterRect(Wnd: HWND): TRect;
begin
  GetWindowRect(hwndTV, Result);
  Result.Left := Result.Right-Result.Left;
  Result.Right := Result.Left + SP_WIDTH;
  Result.Bottom := Result.Bottom-Result.Top;
  Result.Top := 0;
end;

procedure DrawSplitter(Wnd: HWND);
var
  DC: HDC;
  SR: TRect;
  Brsh: HBRUSH;
begin
  DC := GetDC(Wnd);
  SR := GetSplitterRect(Wnd);
  Brsh := CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
  FillRect(DC, SR, Brsh);
  //DrawEdge(DC, SR, BDR_RAISEDOUTER, BF_RECT);
  DeleteObject(Brsh);
  ReleaseDC(Wnd, DC);
end;

{ OnPaint }
function OnPaint(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  DC: HDC;
  lpPaint: PAINTSTRUCT;
  R: TRect;
  SR: TRect;
begin
  Result := 0;
  DC := BeginPaint(Wnd, lpPaint);
  GetClientRect(Wnd, R);
  SR := GetSplitterRect(Wnd);
  DrawSplitter(Wnd);
  //FillRect(DC, SR, GetStockObject(GRAY_BRUSH));
  TextOut(DC, 0, 0, szCaption, lstrlen(szCaption));
  EndPaint(Wnd, lpPaint);
end;

var
  StartX, StartY: Integer;
  Delta: Integer;


{ OnMouseMove }
function OnMouseMove(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  Text: string;
  MR, R, LR: TRect;
  X: Integer;
begin
  Result := 0;
  if wParam and MK_LBUTTON <> 0 then
  begin
    if (Integer(LOWORD(lParam)) < MIN_VALUE) or
       (Integer(LOWORD(lParam)) > MAX_VALUE) then Exit;
    GetClientRect(Wnd, MR);
    GetWindowRect(hwndTV, R);
    MoveWindow(hwndTV, 0, 0, LOWORD(lParam) - Delta, R.Bottom-R.Top, True);
    {SetWindowPos(hwndTV, 0, 0, 0, LOWORD(lParam), R.Bottom-R.Top,
      SWP_NOMOVE or SWP_NOZORDER);}
    TVWidth := LOWORD(lParam) - Delta;
    GetWindowRect(hwndLV, LR);
    {SetWindowPos(hwndLV, 0, LOWORD(lParam) + 4, 0, MR.Right-MR.Left, R.Bottom-R.Top,
      SWP_NOZORDER);}
    X := LOWORD(lParam)+SP_WIDTH-Delta;
    MoveWindow(hwndLV, X, 0, MR.Right-MR.Left-X, R.Bottom-R.Top, True);
    GetClientRect(hwndLV, LR);
    InvalidateRect(hwndLV, @LR, True);
    DrawSplitter(Wnd);
  end;
end;


var
  WndDC: HDC;
  Pen, OldPen: HPEN;

procedure DrawSizingLine(Wnd: HWND; X, Y: Integer);
var
  R: TRect;
begin
  GetClientRect(Wnd, R);
  MoveToEx(WndDC, X, 0, nil);
  LineTo(WndDC, X, R.Bottom);
end;

{ OnMouseMove }
function OnMouseMove2(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
begin
  Result := 0;
  if wParam and MK_LBUTTON <> 0 then
  begin
    DrawSizingLine(Wnd, StartX, StartY);
    StartX := LOWORD(lParam);
    StartY := HIWORD(lParam);
    DrawSizingLine(Wnd, StartX, StartY);
  end;
end;

{ OnLButtonDown }
function OnLButtonDown(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  R: TRect;
begin
  Result := 0;
  SetCapture(Wnd);
  WndDC := GetDCEx(Wnd, 0, DCX_CACHE or DCX_CLIPSIBLINGS or DCX_LOCKWINDOWUPDATE);
  SetROP2(WndDC, R2_XORPEN);
  Pen := CreatePen(PS_DASHDOT, SP_WIDTH, $00FFFFFF);
  OldPen := SelectObject(WndDC, Pen);
  StartX := LOWORD(lParam);
  StartY := HIWORD(lParam);
  GetWindowRect(hwndTV, R);
  Delta := StartX-(R.Right-R.Left);
  //DrawSizingLine(Wnd, StartX, StartY);
end;

{ OnLButtonUp }
function OnLButtonUp(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT;
var
  MR, R: TRect;
  X: Integer;
  EndX: Integer;
begin
  Result := 0;
  ReleaseCapture;
  StartX := LOWORD(lParam);
  StartY := HIWORD(lParam);
  //DrawSizingLine(Wnd, StartX, StartY);
  SelectObject(WndDC, OldPen);
  DeleteObject(Pen);
  SetROP2(WndDC, R2_COPYPEN);
  ReleaseDC(Wnd, WndDC);

  EndX := Integer(LOWORD(lParam));

  if (Integer(LOWORD(lParam)) < MIN_VALUE) then
    EndX := MIN_VALUE;

  if (Integer(LOWORD(lParam)) > MAX_VALUE) then
    EndX := MAX_VALUE;

  GetWindowRect(hwndTV, R);
  TVWidth := EndX - Delta;
  MoveWindow(hwndTV, 0, 0, TVWidth, R.Bottom - R.Top, True);
  InvalidateRect(hwndTV, nil, True);

  GetClientRect(Wnd, MR);
  X := EndX + SP_WIDTH - Delta;
  MoveWindow(hwndLV, X, 0, MR.Right - MR.Left - X, R.Bottom - R.Top, True);
  InvalidateRect(hwndLV, nil, True);

end;

function MainWndProc(Wnd: HWND; uMsg: UINT; wParam: LongInt; lParam: LongInt): LRESULT; stdcall;
begin

  Result := 0;

  case uMsg of

    WM_NOTIFY:
      Result := OnNotify(Wnd, uMsg, wParam, lParam);

    WM_LBUTTONDOWN:
      OnLButtonDown(Wnd, uMsg, wParam, lParam);

    WM_LBUTTONUP:
      OnLButtonUp(Wnd, uMsg, wParam, lParam);

    WM_MOUSEMOVE:
      OnMouseMove(Wnd, uMsg, wParam, lParam);

    WM_SIZE:
      OnSize(Wnd, uMsg, wParam, lParam);

    WM_ERASEBKGND:
      Exit;

    WM_PAINT:
      OnPaint(Wnd, uMsg, wParam, lParam);

    WM_CREATE:
      OnCreate(Wnd, uMsg, wParam, lParam);

    WM_DESTROY:
      begin
        OnDestroy(Wnd, uMsg, wParam, lParam);
        PostQuitMessage(0);
      end;

    WM_CLOSE:
      begin
        OnClose(Wnd, uMsg, wParam, lParam);
        DestroyWindow(Wnd);
      end;

  else
    Result := DefWindowProc(Wnd, uMsg, wParam, lParam);
  end;

end;

function InitInstance(hInst: LongWord): BOOL;
var
  WndClass: WNDCLASSEX;
begin

  with WndClass do
  begin
    WndClass.cbSize := SizeOf(WNDCLASSEX);
    style := CS_DBLCLKS;// or CS_HREDRAW or CS_VREDRAW;
    lpfnWndProc := @MainWndProc;
    cbClsExtra := 0;
    cbWndExtra := 0;
    hInstance := hInst;
    hIcon := LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAINICON));
    hCursor := LoadCursor(0, IDC_SIZEWE);
    hbrBackground := COLOR_BACKGROUND + 1;
    lpszMenuName := nil;
    lpszClassName := szClassName;
    hIconSm := LoadImage(hInst, MakeIntResource(IDI_MAINICON), IMAGE_ICON,
      GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON),
      LR_DEFAULTCOLOR);
  end;

  Result := RegisterClassEx(WndClass) <> 0;
  
end;

function InitApplication(hInst: LongWord; nCmdShow: Integer): BOOL;
begin

  hWndMain := CreateWindowEx(
    0,
    szClassName,
    szCaption,
    WS_OVERLAPPEDWINDOW,
    Integer(CW_USEDEFAULT), Integer(CW_USEDEFAULT),
    Integer(CW_USEDEFAULT), Integer(CW_USEDEFAULT),
    0,
    0,
    hInst,
    nil);

  Result := hWndMain <> 0;

  if Result then
  begin
    ShowWindow(hWndMain, nCmdShow);
    UpdateWindow(hWndmain);
  end;

end;

procedure WinMain(hInst: LongWord; hPrevInst: LongWord);
var
  Msg: TMSG;
begin

  InitCommonControls;

  CoInitialize(nil);

  if not InitInstance(hInst) then
    Halt;

  if not InitApplication(hInst, SW_SHOW) then
    Halt;

  while GetMessage(Msg, 0, 0, 0) do
  begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;

  CoUnInitialize();

  Halt(Msg.wParam);
end;

begin
  WinMain(hInstance, hPrevInst);
end.
